# -*- coding: utf-8 -*-

import os
import os.path
import sys
import xml.dom.minidom

VERSION = "0.1.0"


class BackendDB:
    """ Read and remove content from backenddb.xml file. """
    
    class ConfigItem:
        def __init__(self, entry):
            self.entry = entry
            self.url = entry.getAttribute("url")
            self.data_url = self._read_entry(entry, "conf:data-url")
            self.ini_entry = self._read_entry(entry, "conf:ini-entry")
        
        def _read_entry(self, parent, node_name):
            for child in parent.childNodes:
                if child.nodeType == parent.ELEMENT_NODE and child.nodeName == node_name:
                    for content in child.childNodes:
                        if content.nodeType == child.TEXT_NODE:
                            return content.data
            return ""
        
        def __str__(self):
            return "<ConfigItem, url: {}, data-url: {}, ini-entry: {}>".format(
                    self.url, self.data_url, self.ini_entry)
    
    def __init__(self, dom):
        self.dom = dom
    
    def toxml(self):
        return self.dom.toxml("utf-8")
    
    def config_iter(self):
        """ Iterates configurations. 
            Revoked entries are ignored.
        """
        top_nodes = self.dom.getElementsByTagName("conf:configuration-backend-db")
        for top_node in top_nodes:
            for entry in top_node.childNodes:
                if entry.nodeType == top_node.ELEMENT_NODE and entry.nodeName == "conf:configuration":
                    if entry.getAttribute("revoked") != "true":
                        yield self.ConfigItem(entry)
    
    def backend_item_iter(self):
        """ Iterates configurations which contains conf:data-url. 
            These entries have substituted files in the backend.
        """
        for entry in self.config_iter():
            if entry.data_url:
                yield entry
    
    def remove_entry(self, entry_item):
        """ Removes specified entry from this data entries. 
            @param entry_item item entry
        """
        if entry_item.entry:
            try:
                entry = entry_item.entry
                entry.parentNode.removeChild(entry)
                entry.unlink()
            except Exception as e:
                print(e)
            entry_item.entry = None
    
    def comment_entry(self, entry_item):
        """ Comment out specified entry in this data entries.
            @param entry_item item entry
        """
        if entry_item.entry:
            try:
                entry = entry_item.entry
                comment = self.dom.createComment(entry.toxml("utf-8"))
                entry.parentNode.insertBefore(comment, entry)
                entry.parentNode.removeChild(entry)
                entry.unlink()
            except Exception as e:
                print(e)
            entry_item.entry = None


class ConfigurationDOMHelper:
    """ Helps to obtain some value from XCU file. """
    
    def __init__(self, dom):
        self.dom = dom
    
    def component_data_iter(self, target_package, target_name):
        """ Iterates elements in component data having package and name specified.
            @param target_package name of package
            @param target_name name in the package
        """
        top_nodes = self.dom.getElementsByTagName("oor:component-data")
        for top_node in top_nodes:
            package = top_node.getAttribute("oor:package")
            package = package.strip("/")
            if package == target_package and top_node.getAttribute("oor:name") == target_name:
                yield top_node
    
    def children_iter(self, parent_node, attr_name, attr_value):
        """ Iterates child element nodes which has specified attribute.
            @param parent_node node from dom
            @param attr_name attribute name
            @param attr_value attribute value
        """
        for node in parent_node.childNodes:
            if node.nodeType == parent_node.ELEMENT_NODE and \
               node.getAttribute(attr_name) == attr_value:
                   yield node
    
    def value_getter(self, parent_node):
        """ Returns content of first "value" node. 
            Localization is not supported.
            @param parent_node node from dom
        """
        for node in parent_node.childNodes:
            if node.nodeType == node.ELEMENT_NODE and node.nodeName == "value":
                for child in node.childNodes:
                    if child.nodeType == node.TEXT_NODE:
                        return child.data
    
    def get_properties(self, node):
        """ Returns properties as a map which contains values from specified node.
            @param node node from dom
        """
        props = {}
        for prop in node.childNodes:
            if prop.nodeType == node.ELEMENT_NODE and prop.nodeName == "prop":
                prop_name = prop.getAttribute("oor:name")
                type = prop.getAttribute("oor:type")
                value = self.value_getter(prop)
                # todo, other type of list of data
                if type == "oor:string-list":
                    props[prop_name] = value.split(" ")
                else:
                    props[prop_name] = value
        return props


try:
    from io import StringIO
except:
    from StringIO import StringIO


class Logger:
    # todo, replace with default logger
    
    def __init__(self, verbose):
        self.verbose = verbose
        self.f = StringIO()
    
    def get(self, close=True):
        s = self.f.getvalue()
        if close:
            self.f.close()
        return s
    
    def log(self, message, verbose=False, no_lf=False):
        write = self.verbose if verbose else True
        if write:
            try:
                message = message.decode("utf-8")
            except:
                pass
            self.f.write(message)
            if not no_lf:
                self.f.write(u"\n")
    
    def error(self, message):
        self.log(message, verbose=False)


class FileTools:
    
    def __init__(self):
        pass
    
    def read_text(self, path):
        with open(path, "r") as f:
            return f.read()
    
    def write_text(self, path, data):
        with open(path, "w") as f:
            f.write(data)
    
    def delete_file(self, path):
        os.remove(path)
    
    def file_exists(self, path):
        return os.path.exists(path)
    
    def list_dir_contents(self, path):
        join = os.path.join
        return [join(path, name) for name in os.listdir(path)]
    
    def is_dir(self, path):
        return os.path.isdir(path)


class OfficeFileTools(FileTools):
    
    def __init__(self, ctx):
        """ Initialize file tools for office environment.
            @param ctx component context
        """
        FileTools.__init__(self)
        self.ctx = ctx
        self.sfa = self._create("com.sun.star.ucb.SimpleFileAccess")
    
    def _create(self, name):
        """ Creates new instance of service.
            @param name service name
            @return new instance of the service
        """
        return self.ctx.getServiceManager().createInstanceWithContext(name, self.ctx)
    
    def read_text(self, path):
        """ Read text content from URL. 
            @param path file URL
            @return text content of the file
        """
        text_input = self._create("com.sun.star.io.TextInputStream")
        text_input.setEncoding("UTF-8")
        text_input.setInputStream(self.sfa.openFileRead(path))
        s = text_input.readString((), False)
        text_input.closeInput()
        return s
    
    def write_text(self, path, data):
        """ Writes text content to URL. 
            @param path file URL
            @param data text data
        """
        text_output = self._create("com.sun.star.io.TextOutputStream")
        text_output.setEncoding("UTF-8")
        text_output.setOutputStream(self.sfa.openFileWrite(path))
        text_output.writeString(data)
        text_output.closeOutput()
    
    def delete_file(self, path):
        self.sfa.kill(path)
    
    def file_exists(self, path):
        return self.sfa.exists(path)
    
    def list_dir_contents(self, path):
        return self.sfa.getFolderContents(path, True)
    
    def is_dir(self, path):
        return self.sfa.isFolder(path)


class ConfigurationBackendCleaner:
    """ Cleans configuration backend of extension installation. """
    
    def __init__(self, path, out_dir=None, verbose=False, dryrun=False, checkers=None):
        """ Initialize cleaner.
            @param path path to user's profile
            @param out_dir output directory
            @param verbose verbose output
            @param dryrun only validation
            @return new instance
        """
        self.path = path
        #self.out_dir = out_dir # not supported yet
        self.verbose = verbose
        self.dryrun = dryrun
        if checkers is None:
            checkers = CONFIGURATION_CHECKERS # all checkers
        self.checkers = checkers
        self.ft = FileTools()
        self.config_backend_path = None
        self.db = None
        self.logger = Logger(verbose)
        self.vlog("{} instance created, verbose: {}, dryrun: {}, user's profile: {}".format(
                self.__class__.__name__, self.verbose, self.dryrun, self.path))
    
    def log(self, message):
        self.logger.log(message)
    
    def vlog(self, message):
        self.logger.log(message, verbose=True)
    
    def elog(self, message):
        self.logger.error(message)
    
    def write_log(self):
        print(self.logger.get())
    
    def _get_target_dir(self, user_path):
        if not user_path.endswith("/"):
            user_path = user_path + "/"
        return user_path + "uno_packages/cache/registry/" + \
            "com.sun.star.comp.deployment.configuration.PackageRegistryBackend"
    
    def _get_backenddb_path(self, dir_path):
        if not dir_path.endswith("/"):
            dir_path = dir_path + "/"
        return dir_path + "backenddb.xml"
    
    def _to_path(self, path):
        return path
    
    def _get_short_path(self, path):
        PACKAGES_CACHE = "vnd.sun.star.expand:$UNO_USER_PACKAGES_CACHE"
        BACKEND = "vnd.sun.star.expand:$UNO_USER_PACKAGES_CACHE/registry/com.sun.star.comp.deployment.configuration.PackageRegistryBackend"
        if path.startswith(BACKEND):
            return path.replace(BACKEND, "$BACKEND")
        elif path.startswith(PACKAGES_CACHE):
            return path.replace(PACKAGES_CACHE, "$PACKAGES")
        return path
    
    def _parse_configuration_files(self, base_path):
        broken_entries = []
        self.vlog("Searching files in: " + base_path)
        for item in self.db.backend_item_iter():
            path = item.ini_entry
            if not path.startswith("vnd.sun.star.expand:"):
                path = "vnd.sun.star.expand:" + path
            if self._parse_configuration_file(path):
                broken_entries.append(item)
        
        if broken_entries:
            self.log("Following files seems broken: ")
            for broken in broken_entries:
                self.log(broken.ini_entry)
                # todo, reason?
                #self.db.remove_entry(broken)
                self.db.comment_entry(broken)
                # if no extension file exists, remove this?
            self._write_backenddb(self.config_backend_path)
    
    def _parse_configuration_file(self, path):
        self.vlog("Checking file: " + self._get_short_path(path))
        self.vlog("Reading file content: " + self._get_short_path(path))
        dom = self._create_dom(path)
        
        is_broken = False
        for checker in self.checkers:
            # todo, out_path
            try:
                if not checker(dom, self.ft, 
                                    verbose=self.verbose, dryrun=self.dryrun).check(self, path):
                    is_broken = True
            except NotSupportedError:
                pass
        return is_broken
    
    def _init_backenddb(self, backend_dir):
        db_path = self._get_backenddb_path(backend_dir)
        self.vlog("Loading backend database: " + db_path)
        dom = self._create_dom(db_path)
        self.db = BackendDB(dom)
    
    def _write_backenddb(self, backend_dir):
        db_path = self._get_backenddb_path(backend_dir)
        try:
            if not self.dryrun:
                self.ft.delete_file(db_path)
                self.ft.write_text(db_path, self.db.toxml())
            self.log("Update backend database: " + db_path)
        except Exception as e:
            print(e)
            self.elog("Error while writing: " + db_path)
    
    def _read_text(self, path):
        """ Returns text content from file path.
            @param path path to a file
        """
        try:
            return self.ft.read_text(path)
        except Exception as e:
            # todo, what kind of error?
            self.elog("Error while reading: " + path)
            self.log("Skipping file: " + path)
            raise
    
    def _create_dom(self, path):
        """ Returns constructed dom from path.
            @param path path to a xml file
        """
        contents = self._read_text(path)
        try:
            return xml.dom.minidom.parseString(contents.encode("utf-8"))
        except Exception as e:
            self.elog("Error while parsing as dom: " + path)
            self.log("Skipping file: " + path)
            raise
    
    def start(self):
        """ Start cleanup. """
        config_backend_path = self._get_target_dir(self.path)
        self.config_backend_path = config_backend_path
        self.vlog("Configuration backend: {}".format(config_backend_path))
        
        if not self.ft.file_exists(config_backend_path):
            self.elog("Error", "Directory not found: \n" + config_backend_path)
            return
        
        self.vlog("Dryrun mode will not write any result into files.")
        
        self._init_backenddb(config_backend_path)
        
        self._parse_configuration_files(config_backend_path)
        self.write_log()


class OfficeCleaner(ConfigurationBackendCleaner):
    
    def __init__(self, ctx, path=None, verbose=False, dryrun=False, checkers=None):
        """ Initialize cleaner.
            @param ctx component context
            @param path URL to user's profile. Current profile is used if not specified.
        """
        self.ctx = ctx
        if path is None:
            path = self._get_user_profile_path()
        ConfigurationBackendCleaner.__init__(self, path, verbose=verbose, dryrun=dryrun, checkers=checkers)
        # replace FileTools
        self.ft = OfficeFileTools(self.ctx)
        import uno
        self.fileUrlToSystemPath = uno.fileUrlToSystemPath
    
    def write_log(self):
        log = self.logger.get()
        doc = self._create_writer_document()
        doc.getText().setString(log)
        doc.setModified(False)
    
    def _create_writer_document(self):
        desktop = self._create("com.sun.star.frame.Desktop")
        doc = desktop.loadComponentFromURL("private:factory/swriter", "_blank", 0, ())
        return doc
    
    def _get_user_profile_path(self):
        return self._substitute_url("$(user)")
    
    def _to_path(self, path):
        return self.fileUrlToSystemPath(path)
    
    def msgbox(self, title, message, type=0):
        desktop = self._create("com.sun.star.frame.Desktop")
        toolkit = self._create("com.sun.star.awt.Toolkit")
        msgbox = toolkit.createMessageBox(desktop.getCurrentFrame().getContainerWindow(), 
                    type, 1, title, message)
        return msgbox.execute()
    
    def error(self, title, message):
        return self.msgbox(title, message, type=3)
    
    def _create(self, name):
        """ Creates new instance of service.
            @param name service name
            @return new instance of the service
        """
        return self.ctx.getServiceManager().createInstanceWithContext(name, self.ctx)
    
    def _substitute_url(self, path):
        """ Returns substituted URL. 
            @param path URL which might contain variables
            @return new URL
        """
        return self._create("com.sun.star.util.PathSubstitution").substituteVariables(path, True)


class NotSupportedError(Exception):
    """ This file is not supported by this checker. """
    pass


class Checker:
    """ Base class for checks something. """
    
    def __init__(self, verbose=False, dryrun=True):
        self.verbose = verbose
        self.dryrun = dryrun


class ConfigurationDOMChecker(Checker):
    """ DOM based configuration file checker. """
    
    def __init__(self, dom, file_tools, verbose=False, dryrun=True):
        Checker.__init__(self, verbose, dryrun)
        self.dom = dom
        self.ft = file_tools
    
    def check(self, parent, path):
        """ Checks this configuration.
            @param parent cleaner
            @param path path to file to be checked
        """
        return True # OK, no problem found


class SpellCheckerDictionaryChecker(ConfigurationDOMChecker):
    """ Checks existence of dictionary files in configuration DOM. """
    
    def check(self, parent, out_path):
        parent.vlog("Checking by {}".format(self.__class__.__name__))
        broken_nodes = []
        helper = ConfigurationDOMHelper(self.dom)
        try:
            # find nodes which contains broken Locations node
            for entry in self._data_iter(parent, helper):
                is_broken = False
                entry_name = entry.getAttribute("oor:name")
                props = helper.get_properties(entry)
                parent.vlog("Entry: {}, Format: {}, Locales: {}".format(
                                entry_name, props.get("Format", "unknown"), 
                                " ".join(props.get("Locales", ()))))
                for location in props.get("Locations", ()):
                    # todo, expand URL
                    is_broken = not self.ft.file_exists(location)
                    if is_broken:
                        parent.log("File not found: {}".format(parent._get_short_path(location)))
                    else:
                        parent.vlog("Ok: {}".format(parent._get_short_path(location)))
                    if is_broken:
                        break
                broken_nodes.append(entry if is_broken else None)
        except NotSupportedError:
            raise
        except Exception as e:
            print(e)
            import traceback
            traceback.print_exc()
        
        dom = self.dom
        has_broken_entry = any(broken_nodes)
        if has_broken_entry:
            for entry in broken_nodes:
                if entry:
                    entry_name = entry.getAttribute("oor:name")
                    # put this entry in comment
                    comment = dom.createComment(entry.toxml("utf-8"))
                    parent_node = entry.parentNode
                    parent_node.insertBefore(comment, entry)
                    # remove original
                    parent_node.removeChild(entry)
                    entry.unlink()
                    parent.vlog("Entry commented: {}".format(entry_name))
                    try:
                        if not self.dryrun:
                            # todo, delete_file if contents is going to be shorter
                            self.ft.write_text(out_path, dom.toxml("utf-8"))
                        parent.log("Entry \"{}\" purged from: {}".format(entry_name, parent._get_short_path(out_path)))
                    except Exception as e:
                        self.error("Error while writing: " + out_path)
                        raise
        return not has_broken_entry
    
    def _data_iter(self, parent, helper):
        """ Iterates nodes in this configuration file.
            If org.openoffice.Office.Linguistic data is not find, NotSupportedError is raised.
            @param helper dom helper
        """
        counter = 0
        for top_node in helper.component_data_iter("org.openoffice.Office", "Linguistic"):
            counter += 1
            for entry in self._dictionaries_iter(helper, top_node):
                yield entry
        if counter == 0:
            parent.vlog("This file is not supported by {}.".format(self.__class__.__name__))
            raise NotSupportedError()
    
    def _dictionaries_iter(self, helper, parent_node):
        """ Iterates dictionary entries.
            @param helper dom helper
            @param parent_node dom node
        """
        for service_manager_node in helper.children_iter(parent_node, "oor:name", "ServiceManager"):
            for dictionaries_node in helper.children_iter(service_manager_node, "oor:name", "Dictionaries"):
                # check Locations subnode
                for entry in dictionaries_node.childNodes:
                    if entry.nodeType == dictionaries_node.ELEMENT_NODE and entry.nodeName == "node":
                        yield(entry)


# All configuration based checkers
CONFIGURATION_CHECKERS = (
    SpellCheckerDictionaryChecker,
    
)


def _execute(verbose, dryrun):
    ctx = XSCRIPTCONTEXT.getComponentContext()
    OfficeCleaner(ctx, verbose=verbose, dryrun=dryrun).start()


def check(*args):
    """ Checks configuration backend. """
    _execute(verbose=True, dryrun=True)


def cleanup(*args):
    """ Executes configuration backend cleaner. """
    _execute(verbose=True, dryrun=False)


g_exportedScripts = check, cleanup,


def main():
    # not working yet
    import argparse
    parser = argparse.ArgumentParser(prog_name="config_backend_cleaner",
                            description="Tries to remove broken configuration files from extension backend.")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s " + VERSION,
                        help="Prints program version.")
    parser.add_argument("-b", "--verbose", action="store_true", default=False,
                        help="Verbose output.")
    parser.add_argument("-w", "--write", action="store_true", default=False,
                        help="Try to fix problems in files if found.")
    #parser.add_argument("-o", "--out", type=str, nargs=1, default=None,
    #                    help="Output directory")
    
    args = parser.parse_args()
    Cleaner(verbose=args.verbose, dryrun=(not args.write)).start()

if __name__ == "__main__":
    #main()
    pass
